import datetime as dt
import os
import subprocess
from functools import partial
from typing import Dict, Generator, List

import aiobotocore
import boto3
import pytest
from _pytest.monkeypatch import MonkeyPatch
from aiobotocore.session import AioSession
from fastapi.testclient import TestClient

from examples.app import app
from examples.models import Account, Card, File
from fast_agave.tasks import sqs_tasks


@pytest.fixture
def client() -> Generator[TestClient, None, None]:
    client = TestClient(app)
    yield client


@pytest.fixture
def accounts() -> Generator[List[Account], None, None]:
    user_id = 'US123456789'
    accs = [
        Account(
            name='Frida Kahlo',
            user_id=user_id,
            created_at=dt.datetime(2020, 1, 1),
        ),
        Account(
            name='Sor Juana Inés',
            user_id=user_id,
            created_at=dt.datetime(2020, 2, 1),
        ),
        Account(
            name='Leona Vicario',
            user_id=user_id,
            created_at=dt.datetime(2020, 3, 1),
        ),
        Account(
            name='Remedios Varo',
            user_id='US987654321',
            created_at=dt.datetime(2020, 4, 1),
        ),
    ]

    for acc in accs:
        acc.save()
    yield accs
    for acc in accs:
        acc.delete()


@pytest.fixture
def account(accounts: List[Account]) -> Generator[Account, None, None]:
    yield accounts[0]


@pytest.fixture
def other_account(accounts: List[Account]) -> Generator[Account, None, None]:
    yield accounts[-1]


@pytest.fixture
def files() -> Generator[List[File], None, None]:
    user_id = 'US123456789'
    accs = [
        File(
            name='Frida Kahlo',
            user_id=user_id,
        ),
    ]

    for acc in accs:
        acc.save()
    yield accs
    for acc in accs:
        acc.delete()


@pytest.fixture
def file(files: List[File]) -> Generator[File, None, None]:
    yield files[0]


@pytest.fixture
def cards() -> Generator[List[Card], None, None]:
    user_id = 'US123456789'
    cards = [
        Card(
            number='5434000000000001',
            user_id=user_id,
            created_at=dt.datetime(2020, 1, 1),
        ),
        Card(
            number='5434000000000002',
            user_id=user_id,
            created_at=dt.datetime(2020, 2, 1),
        ),
        Card(
            number='5434000000000003',
            user_id=user_id,
            created_at=dt.datetime(2020, 3, 1),
        ),
        Card(
            number='5434000000000004',
            user_id='US987654321',
            created_at=dt.datetime(2020, 4, 1),
        ),
    ]

    for card in cards:
        card.save()
    yield cards
    for card in cards:
        card.delete()


@pytest.fixture
def card(cards: List[Card]) -> Generator[Card, None, None]:
    yield cards[0]


@pytest.fixture(scope='session')
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_DEFAULT_REGION'] = 'us-east-1'
    boto3.setup_default_session()


@pytest.fixture(scope='session')
def aws_endpoint_urls(
    aws_credentials,
) -> Generator[Dict[str, str], None, None]:
    sqs = subprocess.Popen(['moto_server', 'sqs', '-p', '4000'])

    endpoints = dict(
        sqs='http://127.0.0.1:4000/',
    )
    yield endpoints
    sqs.kill()


@pytest.fixture(autouse=True)
def patch_tasks_count(monkeypatch: MonkeyPatch) -> None:
    def one_loop(*_, **__):
        # Para pruebas solo necesitamos un ciclo
        yield 0

    monkeypatch.setattr(sqs_tasks, 'count', one_loop)


@pytest.fixture(autouse=True)
def patch_create_client(aws_endpoint_urls, monkeypatch: MonkeyPatch) -> None:
    create_client = AioSession.create_client

    def mock_create_client(*args, **kwargs):
        service_name = next(a for a in args if type(a) is str)
        kwargs['endpoint_url'] = aws_endpoint_urls[service_name]

        return create_client(*args, **kwargs)

    monkeypatch.setattr(AioSession, 'create_client', mock_create_client)


@pytest.fixture
async def sqs_client():
    session = aiobotocore.session.get_session()
    async with session.create_client('sqs', 'us-east-1') as sqs:
        await sqs.create_queue(
            QueueName='core.fifo', Attributes={'FifoQueue': 'true'}
        )
        resp = await sqs.get_queue_url(QueueName='core.fifo')
        sqs.send_message = partial(sqs.send_message, QueueUrl=resp['QueueUrl'])
        sqs.receive_message = partial(
            sqs.receive_message, QueueUrl=resp['QueueUrl']
        )
        sqs.queue_url = resp['QueueUrl']
        yield sqs
        await sqs.purge_queue(QueueUrl=resp['QueueUrl'])



__all__ = ['Account', 'app', 'Card', 'File', 'Transaction']


from .accounts import Account
from .base import app
from .cards import Card
from .files import File
from .transactions import Transaction



from cuenca_validations.types import QueryParams
from mongoengine import Q


def generic_query(query: QueryParams) -> Q:
    filters = Q()
    if query.created_before:
        filters &= Q(created_at__lt=query.created_before)
    if query.created_after:
        filters &= Q(created_at__gt=query.created_after)
    exclude_fields = {
        'created_before',
        'created_after',
        'active',
        'limit',
        'page_size',
        'key',
    }
    fields = query.dict(exclude=exclude_fields)
    if 'count' in fields:
        del fields['count']
    return filters & Q(**fields)



import asyncio
import json
from functools import wraps
from itertools import count
from typing import Callable, Coroutine

import aiobotocore


async def run_task(
    coro: Coroutine,
    sqs,
    queue_url: str,
    receipe_handle: str,
) -> None:
    await coro
    await sqs.delete_message(
        QueueUrl=queue_url,
        ReceiptHandle=receipe_handle,
    )


def task(
    queue_url: str,
    region_name: str,
    wait_time_seconds: int = 15,
    visibility_timeout: int = 3600,
):
    def task_builder(task_func: Callable):
        @wraps(task_func)
        async def start_task(*args, **kwargs) -> None:
            session = aiobotocore.session.get_session()
            async with session.create_client('sqs', region_name) as sqs:
                for _ in count():
                    response = await sqs.receive_message(
                        QueueUrl=queue_url,
                        WaitTimeSeconds=wait_time_seconds,
                        VisibilityTimeout=visibility_timeout,
                    )
                    try:
                        messages = response['Messages']
                    except KeyError:
                        continue

                    for message in messages:
                        body = json.loads(message['Body'])
                        asyncio.create_task(
                            run_task(
                                task_func(body),
                                sqs,
                                queue_url,
                                message['ReceiptHandle'],
                            )
                        )

        return start_task

    return task_builder



from unittest.mock import AsyncMock

from _pytest.monkeypatch import MonkeyPatch
from fastapi.testclient import TestClient

from examples.middlewares.authed import AuthedMiddleware
from fast_agave.exc import UnauthorizedError


def test_iam_healthy(client: TestClient) -> None:
    resp = client.get('/')
    assert resp.status_code == 200
    assert resp.json() == dict(greeting="I'm healthy!!!")


def test_cuenca_error_handler(client: TestClient) -> None:
    resp = client.get('/raise_cuenca_errors')
    assert resp.status_code == 401
    assert resp.json() == dict(error='you are not lucky enough!', code=101)


def test_fast_agave_error_handler(client: TestClient) -> None:
    resp = client.get('/raise_fast_agave_errors')
    assert resp.status_code == 401
    assert resp.json() == dict(error='nice try!')


def test_fast_agave_error_handler_from_middleware(
    client: TestClient, monkeypatch: MonkeyPatch
) -> None:
    monkeypatch.setattr(
        AuthedMiddleware,
        'authorize',
        AsyncMock(side_effect=UnauthorizedError('come back to the shadows!')),
    )
    resp = client.get('/you_shall_not_pass')
    assert resp.status_code == 401
    assert resp.json() == dict(error='come back to the shadows!')



import json
from unittest.mock import AsyncMock

import pytest

from fast_agave.tasks.sqs_tasks import task

CORE_QUEUE_REGION = 'us-east-1'


@pytest.mark.asyncio
async def test_execute_tasks(sqs_client) -> None:
    test_message = dict(id='abc123', name='fast-agave')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(return_value=None)
    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(async_mock_function)()
    async_mock_function.assert_called_with(test_message)

    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_not_execute_tasks(sqs_client) -> None:
    async_mock_function = AsyncMock(return_value=None)
    # No escribimos un mensaje en el queue
    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(async_mock_function)()
    async_mock_function.assert_not_called()
    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

version = SourceFileLoader('version', 'fast_agave/version.py').load_module()


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='fast_agave',
    version=version.__version__,
    author='Cuenca',
    author_email='dev@cuenca.com',
    description='Rest api blueprints for fast-api',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/fast-agave',
    packages=find_packages(),
    include_package_data=True,
    package_data=dict(agave=['py.typed']),
    python_requires='>=3.8',
    install_requires=[
        'aiobotocore>=2.1.0,<2.2.0',
        'cuenca-validations>=0.9.4,<1.0.0',
        'fastapi>=0.63.0,<0.69.0',
        'mongoengine-plus>=0.0.2,<1.0.0',
        'starlette-context>=0.3.2,<0.4.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import datetime as dt
from tempfile import TemporaryFile
from unittest.mock import MagicMock, patch
from urllib.parse import urlencode

import pytest
from fastapi.testclient import TestClient

from examples.models import Account, Card, File

USER_ID_FILTER_REQUIRED = (
    'examples.middlewares.AuthedMiddleware.required_user_id'
)


def test_create_resource(client: TestClient) -> None:
    data = dict(name='Doroteo Arango')
    resp = client.post('/accounts', json=data)
    json_body = resp.json()
    model = Account.objects.get(id=json_body['id'])
    assert resp.status_code == 201
    assert model.to_dict() == json_body
    model.delete()


def test_create_resource_bad_request(client: TestClient) -> None:
    data = dict(invalid_field='some value')
    resp = client.post('/accounts', json=data)
    assert resp.status_code == 422


def test_retrieve_resource(client: TestClient, account: Account) -> None:
    resp = client.get(f'/accounts/{account.id}')
    assert resp.status_code == 200
    assert resp.json() == account.to_dict()


@patch(USER_ID_FILTER_REQUIRED, MagicMock(return_value=True))
def test_retrieve_resource_user_id_filter_required(
    client: TestClient, other_account: Account
) -> None:
    resp = client.get(f'/accounts/{other_account.id}')
    assert resp.status_code == 404


def test_retrieve_resource_not_found(client: TestClient) -> None:
    resp = client.get('/accounts/unknown_id')
    assert resp.status_code == 404


def test_update_resource_with_invalid_params(client: TestClient) -> None:
    wrong_params = dict(wrong_param='wrong_value')
    response = client.patch(
        '/accounts/NOT_EXISTS',
        json=wrong_params,
    )
    assert response.status_code == 400


def test_retrieve_custom_method(client: TestClient, card: Card) -> None:
    resp = client.get(f'/cards/{card.id}')
    assert resp.status_code == 200
    assert resp.json()['number'] == '*' * 16


def test_update_resource_that_doesnt_exist(client: TestClient) -> None:
    resp = client.patch(
        '/accounts/5f9b4d0ff8d7255e3cc3c128',
        json=dict(name='Frida'),
    )
    assert resp.status_code == 404


def test_update_resource(client: TestClient, account: Account) -> None:
    resp = client.patch(
        f'/accounts/{account.id}',
        json=dict(name='Maria Felix'),
    )
    account.reload()
    assert resp.json()['name'] == 'Maria Felix'
    assert account.name == 'Maria Felix'
    assert resp.status_code == 200


def test_delete_resource(client: TestClient, account: Account) -> None:
    resp = client.delete(f'/accounts/{account.id}')
    account.reload()
    assert resp.status_code == 200
    assert resp.json()['deactivated_at'] is not None
    assert account.deactivated_at is not None


@pytest.mark.usefixtures('accounts')
def test_query_count_resource(client: TestClient) -> None:
    query_params = dict(count=1, name='Frida Kahlo')
    response = client.get(
        f'/accounts?{urlencode(query_params)}',
    )
    assert response.status_code == 200
    assert response.json()['count'] == 1


@pytest.mark.usefixtures('accounts')
def test_query_all_with_limit(client: TestClient) -> None:
    limit = 2
    query_params = dict(limit=limit)
    response = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = response.json()
    assert response.status_code == 200
    assert len(json_body['items']) == limit
    assert json_body['next_page_uri'] is None


@pytest.mark.usefixtures('accounts')
def test_query_all_resource(client: TestClient) -> None:
    query_params = dict(page_size=2)
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 2

    resp = client.get(json_body['next_page_uri'])
    assert resp.status_code == 200
    assert len(json_body['items']) == 2


@pytest.mark.usefixtures('accounts')
def test_query_all_created_after(client: TestClient) -> None:
    query_params = dict(created_after=dt.datetime(2020, 2, 1).isoformat())
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 2


@pytest.mark.usefixtures('accounts')
@patch(USER_ID_FILTER_REQUIRED, MagicMock(return_value=True))
def test_query_user_id_filter_required(client: TestClient) -> None:
    query_params = dict(page_size=2)
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 2
    assert all(item['user_id'] == 'US123456789' for item in json_body['items'])

    resp = client.get(json_body['next_page_uri'])
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 1
    assert all(item['user_id'] == 'US123456789' for item in json_body['items'])


def test_query_resource_with_invalid_params(client: TestClient) -> None:
    wrong_params = dict(wrong_param='wrong_value')
    response = client.get(f'/accounts?{urlencode(wrong_params)}')
    assert response.status_code == 400


@pytest.mark.usefixtures('cards')
def test_query_custom_method(client: TestClient) -> None:
    query_params = dict(page_size=2)
    resp = client.get(f'/cards?{urlencode(query_params)}')
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 2
    assert all(card['number'] == '*' * 16 for card in json_body['items'])

    resp = client.get(json_body['next_page_uri'])
    json_body = resp.json()
    assert resp.status_code == 200
    assert len(json_body['items']) == 2
    assert all(card['number'] == '*' * 16 for card in json_body['items'])


def test_cannot_query_resource(client: TestClient) -> None:
    query_params = dict(count=1, name='Frida Kahlo')
    response = client.get(
        f'/transactions?{urlencode(query_params)}',
    )
    assert response.status_code == 405


def test_cannot_create_resource(client: TestClient) -> None:
    response = client.post('/transactions', json=dict())
    assert response.status_code == 405
    assert response.json() == dict(error='Method Not Allowed')


def test_cannot_update_resource(client: TestClient) -> None:
    response = client.patch('/transactions', json=dict())
    assert response.status_code == 405
    assert response.json() == dict(error='Method Not Allowed')


def test_cannot_delete_resource(client: TestClient) -> None:
    resp = client.delete('/transactions/TR1234')
    assert resp.status_code == 405
    assert resp.json() == dict(error='Method Not Allowed')


def test_not_found(client: TestClient) -> None:
    resp = client.get('/non-registered-endpoint')
    assert resp.status_code == 404
    assert resp.json() == dict(error='Not Found')


def test_download_resource(client: TestClient, file: File) -> None:
    mimetype = 'application/pdf'
    resp = client.get(f'/files/{file.id}', headers={'Accept': mimetype})
    assert resp.status_code == 200
    assert resp.headers.get('Content-Type') == mimetype


def test_upload_resource(client: TestClient) -> None:
    with TemporaryFile(mode='rb') as f:
        file_body = f.read()
    resp = client.post('/files', files=dict(file=file_body))
    assert resp.status_code == 201
    json = resp.json()
    assert json['name'] == 'file'


def test_upload_resource_with_invalid_form(client: TestClient) -> None:
    wrong_form = dict(another_file=b'Whasaaaaap')
    resp = client.post('/files', files=wrong_form)
    assert resp.status_code == 400



from typing import Dict

from fastapi.responses import JSONResponse as Response
from fast_agave.filters import generic_query

from ..models import Card as CardModel
from ..validators import CardQuery
from .base import app


@app.resource('/cards')
class Card:
    model = CardModel
    query_validator = CardQuery
    get_query_filter = generic_query

    @staticmethod
    async def retrieve(card: CardModel) -> Response:
        data = card.to_dict()
        data['number'] = '*' * 16
        return Response(content=data)

    @staticmethod
    async def query(response: Dict):
        for item in response['items']:
            item['number'] = '*' * 16
        return response
